/* mbed library for 240*320 pixel display TFT based on HX8347D LCD Controller
 * Copyright (c) 2011 Peter Drescher - DC2PD
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */


// fix bmp padding for Bitmap function
// speed up pixel
// 30.12.11 fix cls
// 11.03.12 use DMA to speed up
// 15.03.12 use SSEL for TFT CS to enable DMA Register writes
// 06.04.12 fix SSEL CS problem
// 06.04.12 use direct access to the spi register to speed up the library.
// 11.09.12 switch back to using io pin as cs to avoid problems with SSEL CS.
// 21.09.12 fix Bug in BMP_16
// 11.10.12 patch from Hans Bergles to get SPI1 working again
// 03.02.13 add a switch to switch off DMA use for LPC11U24
// 04.03.13 add support for new Kinetis board
// 25.03.13 fix Bug in bitmap for Kinetis board
// 18.10.13 Better Circle function from Michael Ammann
// 2020-2021 Hacked by JHD

#include "SPI_TFT.h"
#include "mbed.h"
// for fast performance on some STM boards




#define BPP         16                  // Bits per pixel    

#if defined TARGET_LPC1768
#define USE_DMA                     // we use dma to speed up
#define NO_MBED_LIB                 // we write direct to the SPI register to speed up  
#endif

#if defined NO_DMA                      // if LPC1768 user want no DMA
#undef USE_DMA
#endif


//extern Serial pc;
//extern DigitalOut xx;     // debug !!

SPI_TFT::SPI_TFT(PinName mosi, PinName miso, PinName sclk, PinName cs, PinName reset, const char *name)
    : _spi(mosi, miso, sclk), _cs(cs), _reset(reset),GraphicsDisplay(name)
{
    orientation = 0;
    char_x = 0;
    tft_reset();
}

int SPI_TFT::width()
{
    if (orientation == 0 || orientation == 2) return 240;
    else return 320;
}


int SPI_TFT::height()
{
    if (orientation == 0 || orientation == 2) return 320;
    else return 240;
}


void SPI_TFT::set_orientation(unsigned int o)
{
    orientation = o;
    switch (orientation) {
        case 0:
            wr_reg(0x16, 0x08);
            break;
        case 1:
            wr_reg(0x16, 0x68);
            break;
        case 2:
            wr_reg(0x16, 0xC8);
            break;
        case 3:
            wr_reg(0x16, 0xA8);
            break;
    }
    WindowMax();
}


// write command to tft register

void SPI_TFT::wr_cmd(unsigned char cmd)
{
    unsigned short spi_d;
    spi_d =  0x7000 | cmd ;
    _spi.write(spi_d);      // mbed lib

}


// write data to tft register
void SPI_TFT::wr_dat(unsigned char dat)
{
    unsigned short spi_d;
    spi_d =  0x7200 | dat;
    _spi.write(spi_d);      
}



// the HX8347-D controller do not use the MISO (SDO) Signal.
// This is a bug - ?
// A read will return 0 at the moment 

unsigned short SPI_TFT::rd_dat (void)
{
    unsigned short val = 0;

    //val = _spi.write(0x73ff);                /* Dummy read 1           */
    //val   = _spi.write(0x0000);              /* Read D8..D15           */
    return (val);
}

// write to a TFT register
void SPI_TFT::wr_reg (unsigned char reg, unsigned char val)
{
    _spi.lock();
    _spi.format(16,3);  
    _cs=0;
    _reset=0;
    wr_cmd(reg);
    _reset=1;
    wr_dat(val);
    _cs=1;
    _spi.unlock();
}

// read from a TFT register
unsigned short SPI_TFT::rd_reg (unsigned char reg)
{
    wr_cmd(reg);
    return(rd_dat());
}

// setup TFT controller - this is called by constructor
void SPI_TFT::tft_reset()
{
                           // 16 Bit SPI  
    _spi.format(16,3);                 // 16 bit spi mode 3

    _spi.frequency(48000000);          // 48 Mhz SPI clock
    _cs = 1;                           // cs high
                      // end reset
    thread_sleep_for(5);

    /* Start Initial Sequence ----------------------------------------------------*/
    wr_reg(0xEA, 0x00);                 /* Reset Power Control 1                */
    wr_reg(0xEB, 0x20);                 /* Power Control 2                      */
    wr_reg(0xEC, 0x0C);                 /* Power Control 3                      */
    wr_reg(0xED, 0xC4);                 /* Power Control 4                      */
    wr_reg(0xE8, 0x40);                 /* Source OPON_N                        */
    wr_reg(0xE9, 0x38);                 /* Source OPON_I                        */
    wr_reg(0xF1, 0x01);                 /*                                      */
    wr_reg(0xF2, 0x10);                 /*                                      */
    wr_reg(0x27, 0xA3);                 /* Display Control 2                    */

    /* Power On sequence ---------------------------------------------------------*/
    wr_reg(0x1B, 0x1B);                 /* Power Control 2                      */
    wr_reg(0x1A, 0x01);                 /* Power Control 1                      */
    wr_reg(0x24, 0x2F);                 /* Vcom Control 2                       */
    wr_reg(0x25, 0x57);                 /* Vcom Control 3                       */
    wr_reg(0x23, 0x8D);                 /* Vcom Control 1                       */

    /* Gamma settings  -----------------------------------------------------------*/
    wr_reg(0x40,0x00);   //   default setup
    wr_reg(0x41,0x00);   //
    wr_reg(0x42,0x01);   //
    wr_reg(0x43,0x13);   //
    wr_reg(0x44,0x10);   //
    wr_reg(0x45,0x26);   //
    wr_reg(0x46,0x08);   //
    wr_reg(0x47,0x51);   //
    wr_reg(0x48,0x02);   //
    wr_reg(0x49,0x12);   //
    wr_reg(0x4A,0x18);   //
    wr_reg(0x4B,0x19);   //
    wr_reg(0x4C,0x14);   //
    wr_reg(0x50,0x19);   //
    wr_reg(0x51,0x2F);   //
    wr_reg(0x52,0x2C);   //
    wr_reg(0x53,0x3E);   //
    wr_reg(0x54,0x3F);   //
    wr_reg(0x55,0x3F);   //
    wr_reg(0x56,0x2E);   //
    wr_reg(0x57,0x77);   //
    wr_reg(0x58,0x0B);   //
    wr_reg(0x59,0x06);   //
    wr_reg(0x5A,0x07);   //
    wr_reg(0x5B,0x0D);   //
    wr_reg(0x5C,0x1D);   //
    wr_reg(0x5D,0xCC);   //

    /* Power + Osc ---------------------------------------------------------------*/
    wr_reg(0x18, 0x36);                 /* OSC Control 1                        */
    wr_reg(0x19, 0x01);                 /* OSC Control 2                        */
    wr_reg(0x01, 0x00);                 /* Display Mode Control                 */
    wr_reg(0x1F, 0x88);                 /* Power Control 6                      */
    thread_sleep_for(5);                          /* Delay 5 ms                           */
    wr_reg(0x1F, 0x80);                 /* Power Control 6                      */
    thread_sleep_for(5);                         /* Delay 5 ms                           */
    wr_reg(0x1F, 0x90);                 /* Power Control 6                      */
    thread_sleep_for(5);                          /* Delay 5 ms                           */
    wr_reg(0x1F, 0xD0);                 /* Power Control 6                      */
    thread_sleep_for(5);                         /* Delay 5 ms                           */

    wr_reg(0x17, 0x05);                 /* Colmod 16Bit/Pixel                   */

    wr_reg(0x36, 0x00);                 /* Panel Characteristic                 */
    wr_reg(0x28, 0x38);                 /* Display Control 3                    */
    thread_sleep_for(40);
    wr_reg(0x28, 0x3C);                 /* Display Control 3                    */
    switch (orientation) {
        case 0:
            wr_reg(0x16, 0x08);
            break;
        case 2:
            wr_reg(0x16, 0xC8);
            break;
        case 3:
            wr_reg(0x16, 0xA8);
            break;   
         case 1:
         default:
            wr_reg(0x16, 0x68);
            break;   
            
    }

    WindowMax ();
}

// Set one pixel
void SPI_TFT::pixel(int x, int y, int color)
{
    wr_reg(0x03, (x >> 0));
    wr_reg(0x02, (x >> 8));
    wr_reg(0x07, (y >> 0));
    wr_reg(0x06, (y >> 8));
    _spi.lock();
    _cs = 0;
    _reset=0;
    wr_cmd(0x22);
    _reset=1;
 //   _spi.format(8,3);                             // 8 bit Mode 3
 //   _spi.write(SPI_START | SPI_WR | SPI_DATA);    // Write : RS = 1, RW = 0
 //   _spi.format(16,3);                            // switch to 16 bit Mode 3
    #ifdef __DIRECTSPI_H_
        _spi.directWrite(color);
    #else
        _spi.write(color);
    #endif                          // Write D0..D15
    _cs = 1;
    _spi.unlock();
}

// define draw area
void SPI_TFT::window (unsigned int x, unsigned int y, unsigned int w, unsigned int h)
{
    wr_reg(0x03, x );
    wr_reg(0x02, (x >> 8));
    wr_reg(0x05, x+w-1 );
    wr_reg(0x04, ((x+w-1) >> 8));
    wr_reg(0x07,  y );
    wr_reg(0x06, ( y >> 8));
    wr_reg(0x09, ( y+h-1 ));
    wr_reg(0x08, ( (y+h-1) >> 8));
}

// set draw area to max
void SPI_TFT::WindowMax (void)
{
    window (0, 0, width(),  height());
}


// clear screen
void SPI_TFT::cls (void)
{
    fprintf(stderr, "CLS \n\r");
    int pixel = ( width() * height());
    WindowMax();
    _spi.lock();
    _cs = 0;  
    _reset=0;
    wr_cmd(0x22);
    _reset=1;                       // 16 bit SPI  
//    _spi.format(8,3);                             // 8 bit Mode 3
//    _spi.write(SPI_START | SPI_WR | SPI_DATA);    // Write : RS = 1, RW = 0
//    _spi.format(16,3);                            // switch back to 16 bit Mode 3
    unsigned int i;
    for (i = 0; i < ( width() * height()); i++)
    #ifdef __DIRECTSPI_H_
        _spi.directWrite(_background);
    #else
        _spi.write(_background);
    #endif
    _cs = 1;
    _spi.unlock();
}

void SPI_TFT::circle(int x0, int y0, int r, int color)
{  
    int x = -r, y = 0, err = 2-2*r, e2;
    do {
        pixel(x0-x, y0+y,color);
        pixel(x0+x, y0+y,color);
        pixel(x0+x, y0-y,color);
        pixel(x0-x, y0-y,color);
        e2 = err;
        if (e2 <= y) {
            err += ++y*2+1;
            if (-x == y && e2 <= x) e2 = 0;
        }
        if (e2 > x) err += ++x*2+1;
    } while (x <= 0);
        
}
 
void SPI_TFT::fillcircle(int x0, int y0, int r, int color)
{
    int x = -r, y = 0, err = 2-2*r, e2;
    do {
        vline(x0-x, y0-y, y0+y, color);
        vline(x0+x, y0-y, y0+y, color);
        e2 = err;
        if (e2 <= y) {
            err += ++y*2+1;
            if (-x == y && e2 <= x) e2 = 0;
        }
        if (e2 > x) err += ++x*2+1;
    } while (x <= 0);
}


// draw horizontal line
void SPI_TFT::hline(int x0, int x1, int y, int color)
{
    int w;
    w = x1 - x0 + 1;
    window(x0,y,w,1);
    _spi.lock();
    _cs = 0;  
    _reset=0;
    wr_cmd(0x22);
    _reset=1;   
                           // switch back to 16 bit Mode 3
            for (int j=0; j<w; j++) {
                  #ifdef __DIRECTSPI_H_
                    _spi.directWrite(color);
                 #else
                    _spi.write(color);
                 #endif  // one line
            }

    _cs = 1;
    _spi.unlock();
    WindowMax();
    return;
}

// draw vertical line
void SPI_TFT::vline(int x, int y0, int y1, int color)
{
    int h;
    h = y1 - y0 + 1;
    window(x,y0,1,h);
    _spi.lock();
    _cs = 0;  
    _reset=0;
    wr_cmd(0x22);
    _reset=1;   
               // switch to 16 bit Mode 3
    for (int y=0; y<h; y++) {
                      #ifdef __DIRECTSPI_H_
                    _spi.directWrite(color);
                 #else
                    _spi.write(color);
                 #endif  // one line
    }

    _cs = 1;
    _spi.unlock();
    WindowMax();
    return;
}


// draw line
void SPI_TFT::line(int x0, int y0, int x1, int y1, int color)
{
    //WindowMax();
    int   dx = 0, dy = 0;
    int   dx_sym = 0, dy_sym = 0;
    int   dx_x2 = 0, dy_x2 = 0;
    int   di = 0;

    dx = x1-x0;
    dy = y1-y0;

    if (dx == 0) {        /* vertical line */
        if (y1 > y0) vline(x0,y0,y1,color);
        else vline(x0,y1,y0,color);
        return;
    }

    if (dx > 0) {
        dx_sym = 1;
    } else {
        dx_sym = -1;
    }
    if (dy == 0) {        /* horizontal line */
        if (x1 > x0) hline(x0,x1,y0,color);
        else  hline(x1,x0,y0,color);
        return;
    }

    if (dy > 0) {
        dy_sym = 1;
    } else {
        dy_sym = -1;
    }

    dx = dx_sym*dx;
    dy = dy_sym*dy;

    dx_x2 = dx*2;
    dy_x2 = dy*2;

    if (dx >= dy) {
        di = dy_x2 - dx;
        while (x0 != x1) {

            pixel(x0, y0, color);
            x0 += dx_sym;
            if (di<0) {
                di += dy_x2;
            } else {
                di += dy_x2 - dx_x2;
                y0 += dy_sym;
            }
        }
        pixel(x0, y0, color);
    } else {
        di = dx_x2 - dy;
        while (y0 != y1) {
            pixel(x0, y0, color);
            y0 += dy_sym;
            if (di < 0) {
                di += dx_x2;
            } else {
                di += dx_x2 - dy_x2;
                x0 += dx_sym;
            }
        }
        pixel(x0, y0, color);
    }
    return;
}

// draw rect
void SPI_TFT::rect(int x0, int y0, int x1, int y1, int color)
{

    if (x1 > x0) hline(x0,x1,y0,color);
    else  hline(x1,x0,y0,color);

    if (y1 > y0) vline(x0,y0,y1,color);
    else vline(x0,y1,y0,color);

    if (x1 > x0) hline(x0,x1,y1,color);
    else  hline(x1,x0,y1,color);

    if (y1 > y0) vline(x1,y0,y1,color);
    else vline(x1,y1,y0,color);

    return;
}


// fill rect
void SPI_TFT::fillrect(int x0, int y0, int x1, int y1, int color)
{

    int h = y1 - y0 + 1;
    int w = x1 - x0 + 1;
    int pixel = h * w;
    window(x0,y0,w,h);
    _spi.lock();
    _cs = 0;  
    _reset=0;
    wr_cmd(0x22);
    _reset=1;   

    for (int p=0; p<pixel; p++) {
                #ifdef __DIRECTSPI_H_
                    _spi.directWrite(color);
                 #else
                    _spi.write(color);
                 #endif  // one line
    }

    _cs = 1;
    _spi.unlock();
    WindowMax();
    return;
}

// set cursor position
void SPI_TFT::locate(int x, int y)
{
    char_x = x;
    char_y = y;
}


// calculate num of chars in a row
int SPI_TFT::columns()
{
    return width() / font[1];
}

// calculate num of rows on the screen
int SPI_TFT::rows()
{
    return height() / font[2];
}

// print a char on the screen
int SPI_TFT::_putc(int value)
{
    if (value == '\n') {    // new line
        char_x = 0;
        char_y = char_y + font[2];
        if (char_y >= height() - font[2]) {
            char_y = 0;
        }
    } else {
        character(char_x, char_y, value);
    }
    return value;
}

// consrtuct the char out of the font
void SPI_TFT::character(int x, int y, int c)
{
    unsigned int hor,vert,offset,bpl,j,i,b;
    unsigned char* zeichen;
    unsigned char z,w;

    if ((c < 31) || (c > 127)) return;   // test char range

    // read font parameter from start of array
    offset = font[0];                    // bytes / char
    hor = font[1];                       // get hor size of font
    vert = font[2];                      // get vert size of font
    bpl = font[3];                       // bytes per line

    if (char_x + hor > width()) {
        char_x = 0;
        char_y = char_y + vert;
        if (char_y >= height() - font[2]) {
            char_y = 0;
        }
    }
    window(char_x, char_y,hor,vert); // char box
    _spi.lock();
    _cs = 0;  
    _reset=0;
    wr_cmd(0x22);
    _reset=1;   

        zeichen = &font[((c -32) * offset) + 4];    // start of char bitmap
        w = zeichen[0];                             // width of actual char
        for (j=0; j<vert; j++) {  //  vert line
            for (i=0; i<hor; i++) {   //  horz line
                z =  zeichen[bpl * i + ((j & 0xF8) >> 3)+1];
                b = 1 << (j & 0x07);
                if (( z & b ) == 0x00) {

                #ifdef __DIRECTSPI_H_
                    _spi.directWrite(_background);
                 #else
                    _spi.write(_background);
                 #endif  // one line

                } else {
                  #ifdef __DIRECTSPI_H_
                    _spi.directWrite(_foreground);
                 #else
                    _spi.write(_foreground);
                 #endif  // one line
                }
            }
        }
   
    _cs = 1;
    _spi.unlock();
    WindowMax();
    if ((w + 2) < hor) {                   // x offset to next char
        char_x += w + 2;
    } else char_x += hor;
}


void SPI_TFT::set_font(unsigned char* f)
{
    font = f;
}


void SPI_TFT::Bitmap(unsigned int x, unsigned int y, unsigned int w, unsigned int h,unsigned char *bitmap)
{
    unsigned int  j;
    int padd;
  
    unsigned short *bitmap_ptr = (unsigned short *)bitmap;

  
    // the lines are padded to multiple of 4 bytes in a bitmap
    padd = -1;
    do {
        padd ++;
    } while (2*(w + padd)%4 != 0);
    window(x, y, w, h);
    _spi.lock();
    _cs = 0;  
    _reset=0;
    wr_cmd(0x22);
    _reset=1;   

    bitmap_ptr += ((h - 1)* (w + padd));              
    unsigned int i;
    for (j = 0; j < h; j++) {        //Lines
        for (i = 0; i < w; i++) {     // copy pixel data to TFT

                  #ifdef __DIRECTSPI_H_
                    _spi.directWrite(*bitmap_ptr);
                 #else
                    _spi.write(*bitmap_ptr);
                 #endif  // one line
                bitmap_ptr++;
  
        }
        bitmap_ptr -= 2*w;
        bitmap_ptr -= padd;
     }

    _cs = 1;
    _spi.unlock();
    WindowMax();
}

int SPI_TFT::BMP_16(unsigned int x, unsigned int y, const char *Name_BMP)
{
 
#define OffsetPixelWidth    18
#define OffsetPixelHeigh    22
#define OffsetFileSize      34
#define OffsetPixData       10
#define OffsetBPP           28
 
    char filename[50];
    unsigned char BMP_Header[54];
    unsigned short BPP_t;
    int PixelWidth,PixelHeigh;
    unsigned int start_data;
    unsigned int    i,off;
    int padd,j;
    unsigned short *line;
 
    sprintf(&filename[0],"%s",Name_BMP);
    
    FILE *Image = fopen((const char *)&filename[0], "rb");  // open the bmp file
    if (!Image) {
        return(0);      // error file not found !
    }

    fread(&BMP_Header[0],1,54,Image);      // get the BMP Header

    if (BMP_Header[0] != 0x42 || BMP_Header[1] != 0x4D) {  // check magic byte
        fclose(Image);
        return(-1);     // error no BMP file
    }

    BPP_t = BMP_Header[OffsetBPP] + (BMP_Header[OffsetBPP + 1] << 8);
    if (BPP_t != 0x0010) {
        fclose(Image);
        return(-2);     // error no 16 bit BMP
    }
  
    PixelHeigh = BMP_Header[OffsetPixelHeigh] + (BMP_Header[OffsetPixelHeigh + 1] << 8) + (BMP_Header[OffsetPixelHeigh + 2] << 16) + (BMP_Header[OffsetPixelHeigh + 3] << 24);
    PixelWidth = BMP_Header[OffsetPixelWidth] + (BMP_Header[OffsetPixelWidth + 1] << 8) + (BMP_Header[OffsetPixelWidth + 2] << 16) + (BMP_Header[OffsetPixelWidth + 3] << 24);
    if (PixelHeigh > height() + y || PixelWidth > width() + x) {
        fclose(Image);
        return(-3);      // to big
    }

    start_data = BMP_Header[OffsetPixData] + (BMP_Header[OffsetPixData + 1] << 8) + (BMP_Header[OffsetPixData + 2] << 16) + (BMP_Header[OffsetPixData + 3] << 24);

    line = (unsigned short *) malloc (2 * PixelWidth); // we need a buffer for a line
    if (line == NULL) {
        return(-4);         // error no memory
    }
 
    // the bmp lines are padded to multiple of 4 bytes
    padd = -1;
    do {
        padd ++;
    } while ((PixelWidth * 2 + padd)%4 != 0);
    int k,l;

    for ( j = PixelHeigh-1 ; j >=0  ; j-- ) {               //Lines bottom up
       
        off = (j * (PixelWidth  * 2 + padd)) + (start_data);   // start of line
        int erra=fseek(Image, off,SEEK_SET);
        int errb=fread(line,1,PixelWidth * 2,Image);       // read a line - slow !
        window(x, (y+PixelHeigh)-j , PixelWidth ,1);
        _spi.lock();
        _cs=0;
        _reset=0;
        wr_cmd(0x22);
        _reset=1;
        _spi.format(16,3);    
                       // switch to 16 bit Mode 3
        for (l = PixelWidth - 1; l >= 0 ; l--) {        // copy pixel data to TFT
            int out = line[(PixelWidth-1)-l];
                  #ifdef __DIRECTSPI_H_
                    _spi.directWrite(line[(PixelWidth-1)-l]);
                 #else
                    _spi.write(ine[(PixelWidth-1)-l]);
                 #endif  // one line
        }

        _cs=1;
        _spi.unlock();

    }

    free (line);
    fclose(Image);
    WindowMax();
    return(1);
}
 
